#include "global.h"
#include "sound.h"
#include "cpu.h"
#include "crystal.h"

/*

Yamaha YM2149, runs as tested on my MSX1.

TODO: filter out high frequencies (normally inaudible, but can probably be used to create triangle/saw waves at half volume)

game quirks (happen on my MSX too):
 - Athletic land: off-tune artifacts in music, intro song especially
 - Magical Tree: clicks in cloud section background music

*/

#define YM2149_VOLUME 6900

static int lut_volume[32];
static int* ym2149_dac;
static int ym2149_bc=0;
static int ym2149_cc=0;
static int ym2149_address;

/* tone channels */
static int a_period;		static int b_period;		static int c_period;
static int a_period_raw;	static int b_period_raw;	static int c_period_raw;
static int a_disable;		static int b_disable;		static int c_disable;
static int a_noise_disable;	static int b_noise_disable;	static int c_noise_disable;
static int a_envelope;		static int b_envelope;		static int c_envelope;
static int a_sample;		static int b_sample;		static int c_sample;
static int a_vol;		static int b_vol;		static int c_vol;
static int a_output;		static int b_output;		static int c_output;

/* noise */
static int noise_period;
static int noise_period_raw;
static int noise_random;
static int noise_sample;

/* envelope */
static int e_period;
static int e_period_raw;
static int e_count;
static int e_invert;
static int e_hold;
static int e_alt;
static int e_disabled;
static int e_vol;


/* recalculate channel output volume */
#define RECALC(x) x##_output=lut_volume[((x##_sample|x##_disable)&(noise_sample|x##_noise_disable))*x##_vol]

/* clock tone channel */
#define CLOCK(x) \
	if (!--x##_period) { \
		if (x##_period_raw>0x70) { x##_period=x##_period_raw; x##_sample^=1; } \
		else if (x##_period_raw>0) { x##_period=x##_period_raw; x##_sample=1; /* prevent artifacts at high freq */ } \
		else { x##_period=0x10; x##_sample=1; } \
		if (x##_vol) RECALC(x); \
	}

static void __fastcall ym2149_update(void)
{
	const int cpu_cycles=*cpu_get_cycles_ptr();
	
	for (;ym2149_cc>cpu_cycles;ym2149_cc-=crystal->rel_cycle) {
		
		if (!--e_period) {
			/* clock envelope */
			if (e_disabled) {
				e_vol=e_invert;
				e_period=0x7fffff00; /* optimisation hack, doesn't affect accuracy */
			}
			else {
				e_vol=e_count^e_invert;
				
				if (!e_count--) {
					if (e_alt) e_invert^=0x1f;
					e_disabled=e_hold;
					e_count=0x1f;
				}
				
				e_period=e_period_raw?e_period_raw:0x10;
			}
			
			if (a_envelope) { a_vol=e_vol; RECALC(a); }
			if (b_envelope) { b_vol=e_vol; RECALC(b); }
			if (c_envelope) { c_vol=e_vol; RECALC(c); }
		}
		
		if (!--noise_period) {
			/* clock noise */
			if ((noise_random+1)&2) {
				noise_sample^=1;
				if (a_vol) RECALC(a);
				if (b_vol) RECALC(b);
				if (c_vol) RECALC(c);
			}
			
			noise_random=noise_random>>1|((noise_random>>3^noise_random)<<16&0x10000);
			
			noise_period=noise_period_raw;
		}
		
		/* clock tone channels */
		CLOCK(a) CLOCK(b) CLOCK(c)
		
		ym2149_dac[ym2149_bc++]=a_output+b_output+c_output;
	}
}


/* ym2149 registers handling */

/* R0-5: 12 bit tone period, clock divider of 16 (/2 for duty cycle)
 anything below 8 is inaudible, some games, eg. Eggerland, rely on that.
 Related to all:
  - also immediately affects period
  - period actually counts up, not down
	reg:	bits:
	R0,2,4	all	tone period fine, channel a,b,c
	R1,3,5	0-3	tone period coarse, channel a,b,c */
#define _R_PERIOD(x) \
	ym2149_update(); period=x##_period; x##_period+=period_raw-x##_period_raw; \
	if (x##_period<=0) { x##_period=period&1; if (x##_period==0) x##_period=0x10; } \
	x##_period_raw=period_raw
#define R_PERIOD_FINE(x) int period,period_raw=v<<4|(x##_period_raw&0xf000); _R_PERIOD(x)
#define R_PERIOD_COARSE(x) int period,period_raw=(v<<12&0xf000)|(x##_period_raw&0xff0); _R_PERIOD(x)

/* R6: noise period, bits 0-4, clock divider of 16 */
#define R_NOISE() \
	int period,period_raw=v?v<<5&0x3e0:0x20; ym2149_update(); /* only 1 reg, so converting 0 to 1 can be done here */ \
	period=noise_period; noise_period+=period_raw-noise_period_raw; \
	if (noise_period<=0) { noise_period=period&1; if (noise_period==0) noise_period=0x20; } \
	noise_period_raw=period_raw

/* R7: enable/disable channels
	bits:
	0,1,2	disable tone channel a,b,c
	3,4,5	disable noise on channel a,b,c
	6,7	io mode, 0=input, 1=output. 6=port a, 7=port b (don't care here) */
#define R_MIXER() \
	ym2149_update(); a_disable=v&1; b_disable=v>>1&1; c_disable=v>>2&1; \
	a_noise_disable=v>>3&1; b_noise_disable=v>>4&1; c_noise_disable=v>>5&1; \
	if (a_vol) RECALC(a); if (b_vol) RECALC(b); if (c_vol) RECALC(c)

/* R10-12: amplitude (channel a,b,c)
	bits:
	0,1,2,3	amplitude
	4	enable envelope */
#define R_AMP(x) \
	ym2149_update(); v&=0x1f; \
	if ((x##_envelope=v&0x10)) x##_vol=e_vol; \
	else x##_vol=v?v<<1|1:0; RECALC(x)

/* R13,14: 16 bit envelope period, clock divider of 8 (16 on AY8910) (*32/16 for complete envelope cycle)
 (low periods can be used to create triangle and saw waves, though deformed due to nonlinear volume)
	reg:	bits:
	R13	all	envelope period fine
	R14	all	envelope period coarse */
#define _R_EPERIOD() \
	ym2149_update(); period=e_period; e_period+=period_raw-e_period_raw; \
	if (e_period<=0) { e_period=period&1; if (e_period==0) e_period=0x10; } \
	e_period_raw=period_raw
#define R_EPERIOD_FINE() int period,period_raw=v<<4|(e_period_raw&0xff000); _R_EPERIOD()
#define R_EPERIOD_COARSE() int period,period_raw=v<<12|(e_period_raw&0xff0); _R_EPERIOD()

/* R15: envelope shape
	bits:
	0	hold
	1	alternate
	2	attack
	3	continue */
#define R_ESHAPE() \
	ym2149_update(); \
	if (v&8) { e_alt=v&2; e_hold=v&1; } \
	else { e_alt=v&4; e_hold=1; } \
	e_invert=v&4?0x1f:0; \
	e_period&=1; if (e_period==0) e_period=0x10; /* clocks immediately */ \
	e_count=0x1f; e_disabled=0

/* R16: IO port a data, R17: IO port b data */
/* (no handling needed) */

/* write to ym2149 register (see above) */
void __fastcall ym2149_write_data(u8 v)
{
	switch (ym2149_address) {
		
		/* tone period */
		case 0: { R_PERIOD_FINE(a); break; }
		case 1: { R_PERIOD_COARSE(a); break; }
		case 2: { R_PERIOD_FINE(b); break; }
		case 3: { R_PERIOD_COARSE(b); break; }
		case 4: { R_PERIOD_FINE(c); break; }
		case 5: { R_PERIOD_COARSE(c); break; }
		
		/* noise period */
		case 6: { R_NOISE(); break; }
		
		/* mixer */
		case 7: R_MIXER(); break;
		
		/* amplitude */
		case 8: R_AMP(a); break;
		case 9: R_AMP(b); break;
		case 10: R_AMP(c); break;
		
		/* envelope */
		case 11: { R_EPERIOD_FINE(); break; }
		case 12: { R_EPERIOD_COARSE(); break; }
		case 13: R_ESHAPE(); break;
		
		/* i/o */
		case 14: case 15: break;
		
		default: break;
	}
}

void __fastcall ym2149_write_address(u8 v)
{
	ym2149_address=v&0xf;
}


/* --- */
void ym2149_set_buffer_counter(int bc)
{
	if (bc!=0&&bc<ym2149_bc) {
		/* copy leftover to start */
		int i=bc,j=ym2149_bc;
		while (i--) ym2149_dac[i]=ym2149_dac[--j];
	}
	
	ym2149_bc=bc;
}

void ym2149_kill_cycle(void) { ym2149_cc-=crystal->ppu_cycle; }
void ym2149_new_frame(void) { ym2149_cc=*cpu_get_cycles_ptr(); }
void ym2149_end_frame(void) { ym2149_update(); }

void ym2149_init(void)
{
	int i,min;
	double vol=YM2149_VOLUME;
	
	/* steps of 1.5dB (envelope, 3dB for normal amp) */
	i=32;
	while (i--) {
		lut_volume[i]=vol+0.5;
		vol/=1.1885022274370184377301224648922; /* 20log x=1.5 dB -> log x=1.5/20 -> x=10^(1.5/20)=1.1885.. */
	}
	
	/* normalize */
	min=lut_volume[0]; i=32;
	while (i--) lut_volume[i]-=min;
	
	ym2149_dac=sound_create_dac();
	ym2149_bc=ym2149_cc=ym2149_address=0;
	
	/* tone channels */
	a_period=		b_period=		c_period=0x10;
	a_period_raw=		b_period_raw=		c_period_raw=0;
	a_disable=		b_disable=		c_disable=1;
	a_noise_disable=	b_noise_disable=	c_noise_disable=1;
	a_envelope=		b_envelope=		c_envelope=0;
	a_sample=		b_sample=		c_sample=1;
	a_vol=			b_vol=			c_vol=0;
	a_output=		b_output=		c_output=0;
	
	/* noise */
	noise_period=noise_period_raw=0x20;
	noise_random=1;
	noise_sample=0;
	
	/* envelope */
	e_period=0x10;
	e_period_raw=0;
	e_count=0x1e;
	e_invert=0;
	e_hold=1;
	e_alt=0;
	e_disabled=0;
	e_vol=0x1f;
}
